/*
    v0.01 - original

    a tap to demonsrate how to get the full event information.

    Not really much of a tap; really a demonstation of the functionality.

    This code replicates the structure TYPE_TapEvent, and the functions TAP_GetEvent, TAP_GetCurrentEvent,
    TAP_EPG_GetExtInfo.  The structure has the event name/description lengths set to 0x100.

    To use in your tap, simply include these functions and replace the declarations/function calls:
        TYPE_TapEvent        with TYPE_exTapEvent
        TAP_GetEvent,        with exTAP_GetEvent
        TAP_GetCurrentEvent, with exTAP_GetCurrentEvent
        TAP_EPG_GetExtInfo   with exTAP_EPG_GetExtInfo

    Note, the TYPE_PlayInfo structure also uses the shortened event description - needless to say
    that this is the cause of the buffer overflow that can occur when using TAP_Hdd_GetPlayInfo.
    This function uses strcpy() to copy the event description; so if it happens to be >128 bytes
    then it trashes things ... I've not re-implemented this function ... yet.

    This tap contains no firmware patches; but it does access an internal data structure.  This data
    structure is at a hard coded location in each firmware revision.


    Use the 'Recall' key to bring up a menu do demonstrate (exit cancels(no more...), recall propagates
    to next tap/firmware).
*/
#include <tap.h>

//Every TAP application needs unique ID and TAP allocations that have same ID cant execute at same time.
//ID consists of DWORD( 4byte ) and the MSB bit should be set by 1.
//The following is how to use in source codes.
#define ID_USERTAP 0xbdb00201
TAP_ID( ID_USERTAP );

//Every TAP application has to include Program name, Author name, and Program description as following
//manners in source code.
TAP_PROGRAM_NAME("event_fix");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("A TAP to fix the event length");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
#define FW_TAP_IDENT_1204       0x8016fc98
#define FW_TAP_TABLE_1204       0x8016fa98
#define FW_EVENT_TABLE_1204     0x8032e698

#define FW_TAP_IDENT_1205       0x8016fe1c
#define FW_TAP_TABLE_1205       0x8016fc1c
#define FW_EVENT_TABLE_1205     0x8032e818

#define FW_TAP_TABLE_1209       0x8017078c          //the tap table base address; 0x20 bytes per tap
#define FW_TAP_IDENT_1209       0x8017098c          //the current tap id
#define FW_EVENT_TABLE_1209     0x80326c4c          //the base of the internal event table, 0x40 bytes per event

//----------------------------------------------------------------------------
//extended to give 0x100 bytes for the name, description
typedef struct
{
	word	evtId;                  //0x00
	dword	startTime;              //0x04
	dword	endTime;                //0x08
	word	duration;               //0x0c
	byte	runningStatus;          //0x0e
	char	eventName[256];         //0x0f
	char	description[256];       //0x10f    8f
	byte	parentalRating;         //0x20f   10f

	byte	satIdx;                 //0x210   110
	word	svcId;                  //0x212   112
	word	orgNetId;               //0x214   114
	word	tsId;                   //0x216   116
} TYPE_exTapEvent;

//this is the internal data structure used to store events
//it _always_ starts at address FW_EVENT_TABLE
//what are the unknown values ???
//the event_name, extended_event_name are pointers to malloc'ed 512K ram blocks
//note that the extended_event_name is a list of (num_extended) zero terminated (name,description) pairs
//with a total length of extended_length  The final pair is terminated with another 0 (i.e. 2 0's indicate end of list)
typedef struct{
    byte  char00;                   //00    // top 2 bits - valid; lower 5 bits - running status
//    byte  char00a:2;                //00
//    byte  char00b:1;                //00
//    byte  running_status:5;         //00
    byte  char01;                   //01
    byte  duration_hour;            //02
    byte  duration_min;             //03
    word  short04;                  //04    //??? normally 0000
    word  event_id;                 //06
    word  start_date;               //08    //mjd
    byte  start_hour;               //0a
    byte  start_min;                //0b
    word  end_date;                 //0c    //mjd
    byte  end_hour;                 //0e
    byte  end_min;                  //0f
    byte  event_name_length;        //10
    byte  parental_rating;          //11
    word  original_network_id;      //12
    word  transport_stream_id;      //14
    word  service_id;               //16
    word  short18;                  //18    //??? normally 0000
    word  sat_index;                //1a
    word  short1c;                  //1c    //??? normally 0000
    word  short1e;                  //1e    //??? normally 0000
    word  short20;                  //20    //??? normally 0000
    word  short22;                  //22    //??? whats this? - a 'random' number, different for each entry
    word  short24;                  //24    //??? normally 0000
    word  short26;                  //26    //??? whats this? - a 'random' number, different for each entry
    byte  *event_name;              //28    //ptr to event name, event description is offset by event_name_length bytes
    byte  char2c;                   //2c    //01 if extended present else 00
    byte  num_extended;             //2d    //number of extended items
    word  extended_length;          //2e    //extended length
    byte  *extended_event_name;     //30
    word  short34;                  //34    //0001 if extended present else 0000
    word  short36;                  //36    //0001 if extended present else 0000
    word  short38;                  //38    //mjd - date added ?
    word  short3a;                  //3a    //??? normally 0000
    word  short3c;                  //3c    //??? normally 0000
    word  short3e;                  //3e    //??? normally 0000
}type_eventtable;

//----------------------------------------------------------------------------

#define pSysID                      0xa3fffffa
#define ptrfix(p) (p + _tap_startAddr)

extern int _appl_version;

typedef struct {
    int  id;
    char str[0x20];
}type_item;

typedef struct {
    int  id;
    int  visible;
    TYPE_Window win;
    int  rgn;
    char *fixme_titlep;
    type_item item[0x10];
}type_menu;

//----------------------------------------------------------------------------
dword fw_tap_ident;
dword fw_tap_table;
dword fw_event_table;

type_menu menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {3, "xxx",          },
        {0, "Show Main Event",  },
        {1, "Show Sub Event",   },
        {2, "Exit Tap",     },
    },
};


int TAP_Main(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1);
void ShowMenu(type_menu *menu);
void DeleteMenu(type_menu *menu);
void ShowMessage(char* lpMessage, char* lpMessage1, dword dwDelay);

TYPE_exTapEvent *exTAP_GetEvent(int chType, int chNum, int *eventNum);
TYPE_exTapEvent *exTAP_GetCurrentEvent(int chType, int chNum);
byte            *exTAP_EPG_GetExtInfo(TYPE_exTapEvent *tapEvtInfo);

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;
    word SysID = *(word*)pSysID;

    TAP_Print("\nopening %s",__tap_program_name__);

    switch(SysID) {
        case 456:
            break;
        default:
            TAP_Print("\nUnsupported system id - %d", SysID);
            TAP_Print("\nTAP only supports system id 456");
            TAP_Print("\n");
            return 0;
    }
    switch(_appl_version) {
        case 0x1204:
            fw_tap_ident = FW_TAP_IDENT_1204;
            fw_tap_table = FW_TAP_TABLE_1204;
            fw_event_table = FW_EVENT_TABLE_1204;
            break;
        case 0x1205:
            fw_tap_ident = FW_TAP_IDENT_1205;
            fw_tap_table = FW_TAP_TABLE_1205;
            fw_event_table = FW_EVENT_TABLE_1205;
            break;
        case 0x1209:
            fw_tap_ident = FW_TAP_IDENT_1209;
            fw_tap_table = FW_TAP_TABLE_1209;
            fw_event_table = FW_EVENT_TABLE_1209;
            break;
        default:
            TAP_Print("\nUnsupported firmware version - %x", _appl_version);
            TAP_Print("\nTAP only supports firmware versions 5.12.04,5.12.05,5.12.09");
            TAP_Print("\n");
            return 0;
    }

    menu.visible = 0;
    return 1;                 //0=application mode, 1=tsr mode
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    int ret;
    dword state, subState;

    switch (event) {
        case EVT_IDLE:
            break;

        case EVT_KEY:
//TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);

            TAP_GetState(&state, &subState);
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                if (menu.visible) { // handling keys when options window is shown
                    param1 = menu_handler(&menu, param1);
                }
                else {
                    param1 = key_handler(param1);
                }
            }
	        break;
    }
    return param1;
}

int menu_handler(type_menu *menu, dword param1)
{
    int selection;
	int channel_type, channel_number;
	TYPE_exTapEvent *evt_info;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
                	TAP_Channel_GetCurrent(&channel_type, &channel_number);
                	evt_info = exTAP_GetCurrentEvent(channel_type, channel_number);
                    ShowMessage(evt_info->eventName, evt_info->description, 500);
                	TAP_MemFree(evt_info);
				    return 0;
				case 1:
                    TAP_Channel_GetCurrentSubTV(&channel_number);
                	evt_info = exTAP_GetCurrentEvent(SVC_TYPE_Tv, channel_number);
                    ShowMessage(evt_info->eventName, evt_info->description, 500);
                	TAP_MemFree(evt_info);
				    return 0;
				case 2:
                    TAP_Print("\nclosing %s",__tap_program_name__);
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1; // cascades key to other TAP's
        case RKEY_Exit:
            DeleteMenu(menu);
            return 0; // exit without Cascading  (doesn't work without key remapping...)
        default:
	        TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
    		return 0;
    }
    return param1;
}


int key_handler(dword param1)
{
    switch(param1) {
        case RKEY_Recall:                 //menu
            ShowMenu(&menu);
            return 0;
    }
    return param1;
}

//----------------------------------------------------------------------------
void ShowMenu(type_menu *menu)
{
    int i;

	if (!menu->visible) {
	    menu->rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
	    TAP_Win_SetDefaultColor(&menu->win);
	    TAP_Win_Create(&menu->win, menu->rgn, 250, 210, 200, 80, FALSE, FALSE);
        TAP_Win_SetTitle(&menu->win, ptrfix(menu->fixme_titlep), 0, FNT_Size_1622);
	    for (i=1; i<=menu->item[0].id; i++) {
    	    TAP_Win_AddItem(&menu->win, menu->item[i].str);
	    }
	    TAP_Win_SetSelection(&menu->win, 0);
	    menu->visible = 1;
	}
}

void DeleteMenu(type_menu *menu)
{
	if (menu->visible) {
	    TAP_Win_Delete(&menu->win);
    	TAP_Osd_Delete(menu->rgn);
	    menu->visible = 0;
    }
}

void ShowMessage(char* lpMessage, char* lpMessage1, dword dwDelay)
{
	dword rgn;															// stores rgn-handle for osd
	dword w1,w2;														// stores width of message-text
	int i;
	int w = 430;
	char Message2[0x40];

	rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE );					// create rgn-handle
	w1 = TAP_Osd_GetW(lpMessage, 0, FNT_Size_1622 ) + 10;				// calculate width of message
	if (w1 > w) w1 = w;												    // if message is too long
	w2 = TAP_Osd_GetW(lpMessage1, 0, FNT_Size_1622 ) + 10;              // calculate width of message

	TAP_Osd_FillBox(rgn, (720-w)/2-5, 265, w+10, 54+22*(strlen(lpMessage1)/50), RGB(19,19,19) );	// draw background-box for border

	TAP_Osd_PutS(rgn, (720-w)/2, 270, (720+w)/2, lpMessage,
    	COLOR_White, RGB(3,5,10), 0, FNT_Size_1622, FALSE, ALIGN_CENTER);	 // show 1. message

    for(i=0; i<strlen(lpMessage1); i+=50) {
        memcpy(Message2, &lpMessage1[i], 50);
        Message2[50] = '\0';
    	TAP_Osd_PutS(rgn, (720-w)/2, 270+22*((i+50)/50), (720+w)/2, Message2,
	    	COLOR_White, RGB(3,5,10), 0, FNT_Size_1622, FALSE, ALIGN_LEFT);	 // show 2. message
	}
	TAP_Delay(dwDelay);													// show it for dwDelay /100 seconds
	TAP_Osd_Delete(rgn);												// release rgn-handle
}
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
TYPE_exTapEvent *exTAP_GetEvent(int chType, int chNum, int *eventNum)
{
    int i,j;
    TYPE_TapChInfo chInfo;
    int num_events;
    type_eventtable *et;
    TYPE_exTapEvent *evInfo;
    char *description;
    int description_length;

    TAP_Channel_GetInfo(chType, chNum, &chInfo);
    num_events = 0;
    for (i=0; i<5000; i++) {
        et = (type_eventtable *)(fw_event_table + 0x40*i);

        //match orgNetId, tsId, svcId
        if (chInfo.svcId == et->service_id) {
            if (chInfo.tsId == et->transport_stream_id) {
                if (chInfo.orgNetId == et->original_network_id) {
                    num_events ++;
                }
            }
        }
    }
    evInfo = TAP_MemAlloc(num_events * sizeof(TYPE_exTapEvent));

    j = 0;
    for (i=0; i<5000; i++) {
        et = (type_eventtable *)(fw_event_table + 0x40*i);

        //match satIdx, orgNetId, tsId, svcId
        if (chInfo.svcId == et->service_id) {
            if (chInfo.tsId == et->transport_stream_id) {
                if (chInfo.orgNetId == et->original_network_id) {
                    if (chInfo.satIdx == et->sat_index) {
                        evInfo[j].evtId = et->event_id;
                        evInfo[j].startTime = (et->start_date << 16) | (et->start_hour << 8) | et->start_min;
                        evInfo[j].endTime = (et->end_date << 16) | (et->end_hour << 8) | et->end_min;
                        evInfo[j].duration = (et->duration_hour << 8) | et->duration_min;
                        evInfo[j].runningStatus = et->char00 & 0x1f;

                        memcpy(evInfo[j].eventName, et->event_name, et->event_name_length);
                        evInfo[j].eventName[et->event_name_length] = '\0';

                        description = (char *)((dword)et->event_name + et->event_name_length);
                        description_length = strlen(description);
                        memcpy(evInfo[j].description, description, description_length);
                        evInfo[j].description[description_length] = '\0';

                        evInfo[j].parentalRating = et->parental_rating;
                        evInfo[j].satIdx = et->sat_index;
                        evInfo[j].svcId = et->service_id;
                        evInfo[j].orgNetId = et->original_network_id;
                        evInfo[j].tsId = et->transport_stream_id;
                        j ++;
                    }
                }
            }
        }
    }

    *eventNum = num_events;
    return evInfo;
}

TYPE_exTapEvent *exTAP_GetCurrentEvent(int chType, int chNum)
{
    int i;
    TYPE_TapChInfo chInfo;
    type_eventtable *et;
    TYPE_exTapEvent *evInfo;
    char *description;
    int description_length;

    TAP_Channel_GetInfo(chType, chNum, &chInfo);

    evInfo = TAP_MemAlloc(sizeof(TYPE_exTapEvent));

    for (i=0; i<5000; i++) {
        et = (type_eventtable *)(fw_event_table + 0x40*i);

        //match satIdx, orgNetId, tsId, svcId
        if ((et->char00 & 0x1f) == 4) {   //running
            if (chInfo.svcId == et->service_id) {
                if (chInfo.tsId == et->transport_stream_id) {
                    if (chInfo.orgNetId == et->original_network_id) {
                        if (chInfo.satIdx == et->sat_index) {
                            evInfo->evtId = et->event_id;
                            evInfo->startTime = (et->start_date << 16) | (et->start_hour << 8) | et->start_min;
                            evInfo->endTime = (et->end_date << 16) | (et->end_hour << 8) | et->end_min;
                            evInfo->duration = (et->duration_hour << 8) | et->duration_min;
                            evInfo->runningStatus = et->char00 & 0x1f;

                            memcpy(evInfo->eventName, et->event_name, et->event_name_length);
                            evInfo->eventName[et->event_name_length] = '\0';

                            description = (char *)((dword)et->event_name + et->event_name_length);
                            description_length = strlen(description);
                            memcpy(evInfo->description, description, description_length);
                            evInfo->description[description_length] = '\0';

                            evInfo->parentalRating = et->parental_rating;
                            evInfo->satIdx = et->sat_index;
                            evInfo->svcId = et->service_id;
                            evInfo->orgNetId = et->original_network_id;
                            evInfo->tsId = et->transport_stream_id;
                            break;
                        }
                    }
                }
            }
        }
    }

    return evInfo;
}

byte *exTAP_EPG_GetExtInfo(TYPE_exTapEvent *tapEvtInfo)
{
    byte *extInfo;
    int i;
    type_eventtable *et;
    TYPE_exTapEvent *evInfo;

    for (i=0; i<5000; i++) {
        et = (type_eventtable *)(fw_event_table + 0x40*i);

        //match satIdx, orgNetId, tsId, svcId, evtId,
        if (tapEvtInfo->evtId == et->event_id) {
            if (tapEvtInfo->svcId == et->service_id) {
                if (tapEvtInfo->tsId == et->transport_stream_id) {
                    if (tapEvtInfo->orgNetId == et->original_network_id) {
                        if (tapEvtInfo->satIdx == et->sat_index) {
//TAP_Print("\nexTAP_EPG_GetExtInfo(), num_extended        = %02x", et->num_extended);
//TAP_Print("\nexTAP_EPG_GetExtInfo(), extended_length     = %04x", et->extended_length);
                            extInfo = TAP_MemAlloc(et->extended_length+1);
                            memcpy(extInfo, et->extended_event_name, et->extended_length);
                            extInfo[et->extended_length] = '\0';
                            break;
                        }
                    }
                }
            }
        }
    }

    return extInfo;
}
//----------------------------------------------------------------------------
